/**
 * @Author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: Release 1.1 $date:  June 17, 1999$
 ** (c) 1999 by Ben Yanbin Zhao and the Regents of the University of California
 ** <P>
 ** The RMI Service that runs a XSet server in the background, accepting
 ** remote/local RMI calls through the XSetServiceIF interface.  This interface
 ** has been updated and greatly expanded since the first 1.0 release.<BR>
 ** This is a simple class, meant only as an RMI wrapper for the SETserver
 ** object.  Programmers can easily take the SETserver object and access the
 ** methods directly.
 **/

package xset;

import java.io.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import org.w3c.dom.*;
import com.ibm.xml.parsers.DOMParser;
import org.xml.sax.InputSource;

/**
 ** Class that makes the SETserver an RMI service
 ** It offers methods to add and delete files from the XSet search engine, 
 ** and to query the dataset in different ways.
 ** <P>
 ** The service is originally empty, and files can be added directory
 ** by directory by calling AddFiles(path)
 **/
public class XSetService extends UnicastRemoteObject implements XSetServiceIF {
  private SETserver server;
  DOMParser parser;
  
  /**
   ** Constructor, just creates an SETserver object
   **/
  public XSetService() throws RemoteException {
    server = new SETserver();
    parser = new DOMParser();
  }
  
  public static void usage() {
    System.out.println("Correct invocation syntax is:");
    System.out.println("  java xset.XSetService &");
  }
  
  public static void main(String argv[]) {
    if (argv.length > 0) {
      usage();
      System.exit(0);
    }
    try {
      XSetService xs = new XSetService();
      Naming.bind("XSet", xs);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  
  /**
   ** Method used to add a single document to the XSet search engine index
   ** @param xmlstring the document to be added
   ** @exception RemoteException as a remote service should
   **/
  public void AddDocument(String xmlstring) 
    throws java.rmi.RemoteException {
    try {
      server.SETregister(xmlstring);
    } catch (XSetException xse) {
      throw new java.rmi.RemoteException(xse.getMessage());
    }
  }
  
  /**
   ** Method used to add files to the XSet search engine index
   ** @param path String that encodes the full path of the directory
   ** that contains the files to be indexed
   ** @exception RemoteException as demanded by RMI
   **/
  public void AddFiles(String path) throws java.rmi.RemoteException {
    System.err.println("AddFiles request received!");
    File dir = new File(path);
    String [] xmlFiles = null;
    Document d = null;
    
    if ((xmlFiles = dir.list(new ExtensionFilter("xml"))) == null) {
      System.err.println("Directory not found or inaccessible!");
    }
    int i = 0;
    try {
      parser.reset();
      for (i = 0; i < xmlFiles.length; i++) {
	System.err.println("File: "+xmlFiles[i]+ " loaded.");
	parser.parse(path+"/"+xmlFiles[i]);
	d = parser.getDocument();
	if (d != null) server.SETregister(d);
	else System.err.println("File failed to parse: "+
				xmlFiles[i] +", ignored.");
      }
    } catch (FileNotFoundException fnfe) {
      fnfe.printStackTrace(System.err);
      throw new RemoteException("Query file not found!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new RemoteException("Parsing ERROR on: "+xmlFiles[i]);
    }
  }
  
  /**
   ** Method to add a single file to the XSet dataset
   ** @param filename String encoding of the filename including full path
   ** @exception RemoteException if file not found or error in IO.
   **/
  public void AddFile(String filename) throws java.rmi.RemoteException {
    System.err.println("AddFile request received!");
    try {
      parser.reset();
      parser.parse(filename);
      Document d = parser.getDocument();
      if (d != null) server.SETregister(d);
      else throw new java.rmi.RemoteException("Parsing ERROR on: "+filename);
    } catch (FileNotFoundException fnfe) {
      fnfe.printStackTrace(System.err);
      throw new RemoteException("Query file not found!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new RemoteException("Parsing ERROR on: "+filename);
    }
  }
  
  /**
   ** Method to clear the XSet index, in reality it just creates
   ** a new instance of the SETserver object
   **/
  public void ClearAll() throws java.rmi.RemoteException {
    System.err.println("ClearAll request received!");
    server = new SETserver();
    System.err.println("XSet cleared.");
  }

  /**
   ** Method to query the XSet server
   ** @param xmlQuery is the query in the form of a text string encoded xml
   ** document.  The query must be well-formed, but not contain a reference
   ** to any DTDs
   ** @return Vector of ServiceEntry objects, each wrapped around a XML Document
   **/
  public Vector QuerySE(String xmlQuery) throws java.rmi.RemoteException {
    System.err.println("Query request received!");
    parser.reset();
    Document d;
    Vector results = null;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xmlQuery.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null) results = server.querySET(d);
      else throw new java.rmi.RemoteException("Query XML parse error!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      System.err.println("Error parsing query string: \n"+xmlQuery);
      results = new Vector();
    }
    return results;
  }
  
  /**
   ** Method that queries the XSet server and returns vector of XML documents
   ** @param xmlQuery is the query in string encoded xml format, which must be well-formed, and should not contain a reference to any DTDs
   ** @return Vector of XML documents in string form
   **/
  public Vector Query(String xmlQuery) throws java.rmi.RemoteException {
    Vector results = QuerySE(xmlQuery);
    //Results will not be null. If no match, then empty vector will be returned
    if (results == null) {
      System.err.println("ERROR: null results returned!!");
      return new Vector();
    }
    if (results.size() == 0) return results;
    Vector xml = new Vector();
    Enumeration e = results.elements();
    ServiceEntry se;
    while (e.hasMoreElements()) {
      se = (ServiceEntry) e.nextElement();
      xml.addElement(se.getXML());
    }
    System.out.println(xml.size());
    //    System.out.println(xml);
    return xml;
  }
  
  /********************************
   *** Query (String, Vector) ***
   *********************************
   ** Method to query the XSet server, and return only those fields requested
   ** <P>
   ** This returned data structure is rather complicated.  It's a vector of
   ** hashtables, one hashtable for each matching XML document found. For
   ** each of these matches, the hashtable hashes each of the requested
   ** tag names to a vector.  The vector contains all the string values
   ** which occur in that tagname in the currently matched XML document.
   ** So it's a Vector of Hashtables of Vectors.
   ** <P>
   ** @param xmlQuery, the xml query in the form of a string encoded XML
   ** @param tagVector, a vector containing String names of requested tags
   ** @return Vector of hashtables that map each requested tag name to its value
   **/
  public Vector Query(String xmlQuery, Vector tagVector) 
    throws java.rmi.RemoteException {
    
    System.err.println("Query request received!");
    parser.reset();
    Document d;
    Vector results = null;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xmlQuery.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null) results = server.querySET(d, tagVector);
      else throw new java.rmi.RemoteException("Query XML parse error!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      System.err.println("Error parsing query string: \n"+xmlQuery);
      results = new Vector();
    }
    return results;
  }

  /*********************************
   ** EnumerateTag (String []) ***
   **********************************
   ** Method that takes in a path to a subpath down to a tag in the form of an
   ** array, and returns an enumeration of the existing values in the treap for
   ** that tag.  Just a simple wrapper around SETserver.enumerateTag()
   ** @param pathArray - the string array of tag names in the path to the tag
   ** @return Vector of unique existing values that exist in the dataset
   **/
  public Vector EnumerateTag(String [] pathArray) 
    throws java.rmi.RemoteException {
    Enumeration e;
    try {
      e = server.enumerateTag(pathArray);
    } catch (XSetException xse) {
      throw new RemoteException(xse.getMessage());
    }
    Vector newV = new Vector();
    while (e.hasMoreElements())
      newV.addElement(e.nextElement());
    return newV;
  }
  
  /********************************
   ** RemoveDocument(Document) **
   *********************************
   ** The method that removes a document from the index structure
   ** @param d The org.w3c.dom document that is to be removed
   **/
  public void RemoveDocument(Document d) throws java.rmi.RemoteException {
    System.err.println("RemoveDocument (Document) request received!");
    try {
      server.SETderegister(d);
    } catch (XSetException xse) {
      throw new java.rmi.RemoteException(xse.getMessage());
    }
  }

  /*******************************
   *** RemoveDocument(String) **
   ********************************
   ** The method that takes in a XML string, and tries to remove it from the
   ** index if it exists
   ** @param xml The XML string representing the XML document to be removed
   **/
  public void RemoveDocument(String xml) throws java.rmi.RemoteException {
    System.err.println("RemoveDocument (String) request received!");
    parser.reset();
    Document d;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xml.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null)  server.SETderegister(d);
      else throw new java.rmi.RemoteException("Parsing error on XML String!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new java.rmi.RemoteException(e.getMessage());
    }
  }
}
